#pragma once
/*

Restrictions:
- Not completely thread safe.
- Doesn't work with producers providing multiple interfaces, this class only deals with the first interface
a producer provides.
- Only one data stream per device is supported. Image data is always retrieved from the first data stream
*/

// Prior to Visual Studio 2015 Update 3, these hooks were non-const. By defining
// the macro DELAYIMP_INSECURE_WRITABLE_HOOKS own non-const definitions of the
// hooks can be provided.
#ifndef DELAYIMP_INSECURE_WRITABLE_HOOKS
#define DELAYIMP_INSECURE_WRITABLE_HOOKS
#endif

#include <string>
#include <list>
#include <set>
#include <map>
#include <stdexcept>
#include <cassert>
#include <fstream>
#include <memory>
#include <algorithm>
#include <functional>
#include "Platform.h"
#include "ProducerProxy.h"
#include "BufferAllocator.h"
#include "GrabResources.h"
#include "WaitObject.h"

#include <GenTL/GenTL_v1_5.h>
#include <GenTL/PFNC.h>

#if defined (CIH_WIN_BUILD)
// Link the import libraries for the GenICam GenApi libraries used for configuring the camera device.
// Please note, that it is required to delay load these DLLs. ("Properties->Linker->Input->Delay Loaded Dlls"
// resp. /DELAYLOAD linker option).
#  pragma comment (lib, LIB_NAME( "GCBase"))
#  pragma comment (lib, LIB_NAME( "GenApi"))
#  pragma comment (lib, "delayimp.lib" )
#  include <delayimp.h>
#  include <cctype>


namespace
{
    extern "C"
    {
        CIH_WEAK_SYMBOL(PfnDliHook   __pfnDliNotifyHook2);
    }

    FARPROC WINAPI DelayHook(unsigned, PDelayLoadInfo);

    PfnDliHook InitHooks()
    {
        PfnDliHook previous = __pfnDliNotifyHook2;
        __pfnDliNotifyHook2 = DelayHook;
        return previous;
    }

    PfnDliHook OldHook = InitHooks();

    // Make sure that the Basler GenApi libraries are found in our installation.
    FARPROC WINAPI DelayHook(unsigned dliNotify, PDelayLoadInfo pdli)
    {
        using namespace GenTLConsumerImplHelper::Impl;
        HMODULE hModule = 0;
        if (dliNotify == dliNotePreLoadLibrary)
        {
            std::string DllName(pdli->szDll);
            std::transform(DllName.begin(), DllName.end(), DllName.begin(),
                [](unsigned char c) { return (unsigned char)std::tolower(c); });

            if (DllName.find("basler") != std::string::npos)
            {
#pragma warning( push, 1 )
#pragma warning( disable: 4996 )
#if CIH_WORDSIZE == 64
                const char* szBaslerTofBinDir = getenv("BASLER_BLAZE_BIN64");
#else
                const char* szBaslerTofBinDir = getenv("BASLER_BLAZE_BIN");
#endif
#pragma warning( pop )

                if (szBaslerTofBinDir != nullptr)
                {
                    std::string binDir(szBaslerTofBinDir);
                    if (*binDir.rbegin() != '\\' && *binDir.rbegin() != '/')
                        binDir += "\\";

                    std::string dllPath = binDir + pdli->szDll;
                    hModule = ::LoadLibraryExA(dllPath.c_str(), nullptr, LOAD_WITH_ALTERED_SEARCH_PATH);
                    if (hModule != 0)
                    {
                        return (FARPROC)hModule;
                    }
                }
            }
            if (!Helper::GetGenicamGenTLPath().empty())
            {
                // Iterate over all entries in GENICAM_GENTL_PATH and try to load the desired module from there
                const char* delimiter = ";";
                std::list<std::string> lstPathEntries = Helper::Tokenize(Helper::GetGenicamGenTLPath(), delimiter);
                for (std::list<std::string>::const_iterator it = lstPathEntries.begin(); it != lstPathEntries.end(); ++it)
                {
                    std::string dllPath = *it;
                    if (*dllPath.rbegin() != '\\' && *dllPath.rbegin() != '/')
                        dllPath += '\\';;
                    dllPath += DllName;
                    hModule = ::LoadLibraryExA(dllPath.c_str(), nullptr, LOAD_WITH_ALTERED_SEARCH_PATH);
                    if (hModule != 0)
                    {
                        return (FARPROC)hModule;
                    }
                }
            }
        }
        // Let either a previously installed hook or the delay load helper function handle all other cases.
        return OldHook != nullptr ? OldHook(dliNotify, pdli) : nullptr;
    }
}
#endif

#define CIH_CHECK_INITIALIZED() \
    if ( ! s_Producer.IsIntialized() )\
    throw LOGICAL_ERROR_EXCEPTION("GenTL producer must be initialized first.")

#define CIH_CHECK_OPEN() CIH_CHECK_INITIALIZED(); if ( ! IsOpen() ) throw LOGICAL_ERROR_EXCEPTION("Camera must be opened first.")

#define CIH_CHECK_ACQ_PREPARED() if ( ! IsAcquisitionPrepared() ) throw LOGICAL_ERROR_EXCEPTION("Acquisition must be prepared first.");

#define CIH_CHECK_NOT_ACQ_PREPARED() if ( IsAcquisitionPrepared() ) throw LOGICAL_ERROR_EXCEPTION("Acquisition already prepared. Call FinishAcquisition() first.");

#define CIH_CHECK_NO_CONTINUOUS_GRAB() if ( IsGrabbing() ) throw LOGICAL_ERROR_EXCEPTION("Continuous grabbing already has been started. Stop it first.");

namespace GenTLConsumerImplHelper
{
    namespace Internal
    {
        template<const char ProducerName[]> struct GrabFinalizer;
        template<const char ProducerName[]> class CPort;
        typedef GenApi::CLock Lock;


        class LockGuard
        {
        public:

            LockGuard()
                : m_lock(nullptr)
            {

            }

            LockGuard(Lock& lock)
                : m_lock(nullptr)
            {
                acquire(lock);
            }

            void acquire(Lock& lock)
            {
                if (m_lock != nullptr)
                {
                    release();
                }
                m_lock = &lock;
                m_lock->Lock();
            }

            void release()
            {
                if (m_lock != nullptr)
                {
                    m_lock->Unlock();
                    m_lock = nullptr;
                }
            }

            ~LockGuard()
            {
                release();
            }

        private:

            LockGuard& operator=(const LockGuard&);
            LockGuard(const LockGuard&);

            Lock* m_lock;
        };
    }

    template <const char ProducerName[]> class CGenTLCamera;

    enum CameraInfoKey
    {
        DeviceID,
        ModelName,
        DisplayName,
        SerialNumber,
        UserDefinedName,
        IpAddress
    };

    struct CameraInfo
    {
        std::string     strDeviceID;
        std::string     strModelName;
        std::string     strDisplayName;
        std::string     strSerialNumber;
        std::string     strUserDefinedName;
        std::string     strIpAddress;

        static const std::string notAvailable;

        CameraInfo()
            : strDeviceID(notAvailable)
            , strModelName(notAvailable)
            , strDisplayName(notAvailable)
            , strSerialNumber(notAvailable)
            , strUserDefinedName(notAvailable)
            , strIpAddress(notAvailable)
        {}

        std::string GetProperty(CameraInfoKey key) const
        {
            std::string strValue;
            switch (key)
            {
            case DeviceID:
                strValue = strDeviceID; break;
            case ModelName:
                strValue = strModelName; break;
            case DisplayName:
                strValue = strDisplayName; break;
            case SerialNumber:
                strValue = strSerialNumber; break;
            case UserDefinedName:
                strValue = strUserDefinedName; break;
            case IpAddress:
                strValue = strIpAddress; break;
            default:
                assert(false && "invalid key");
                strValue = "N/A";
            }
            return strValue;
        }

        bool IsValid() const { return 0 != strDeviceID.compare(notAvailable); }
    };

    CIH_WEAK_SYMBOL(const std::string CameraInfo::notAvailable) ("N/A");

    struct GrabResult
    {
        GrabResult()
        {
            reset();
        }

        void reset()
        {
            pImageData = 0;
            payloadSize = 0;
            payloadType = GenTL::PAYLOAD_TYPE_UNKNOWN;
            pixelFormat = 0;
            hBuffer = GENTL_INVALID_HANDLE;
            frameID = 0;
            timeStamp = 0;
            status = Ok;
        }

        enum GrabResultStatus
        {
            Ok,
            Failed,
            Timeout,
            Stopped,
        };

        void*                   pImageData;
        size_t                  payloadSize;
        size_t                  payloadType;  // see GenTL::PAYLOADTYPE_INFO_IDS
        int64_t                 pixelFormat;  // see PFNC.h
        GenTL::BUFFER_HANDLE    hBuffer;
        uint64_t                frameID;
        uint64_t                timeStamp;
        GrabResultStatus        status;
    };

    namespace Internal
    {
        // The GrabResultPtrRef class is used under the hood for type conversion purposes.
        struct GrabResultPtrRef
        {
            explicit GrabResultPtrRef(GrabResult* pGrabResult, BufferAllocator* pAllocator)
                : m_pGrabResult(pGrabResult)
                , m_pAllocator(pAllocator)
            {
            }

            GrabResult          *m_pGrabResult;
            BufferAllocator     *m_pAllocator;
        };
    }
    // Auto pointer like wrapper for for a GrabResult.
    // When the smart pointer is destroyed, it uses the allocator passed to the constructor
    // to free the memory buffer the GrabResult is pointing to.
    // !!!ATTENTION!!!: If a user provided allocator has been provided to the GenTL camera, client code must ensure
    // that the allocator is NOT destroyed UNTIL all GrabResultPtr referencing that allocator
    // are destroyed! In case the ownership of a user provided allocator is passed to a GenTLCamera,
    // the camera object must live until all GrabResultPtrs are destroyed (e.g. by leaving the scope
    // or explicitly calling the GrabResultPtr::Reset() method.
    class GrabResultPtr
    {
    public:

        GrabResultPtr()
            : m_pAllocator(nullptr)
            , m_pGrabResult(nullptr)
        {}

        explicit GrabResultPtr(GrabResult& grabResult, BufferAllocator* pAllocator, bool clone = false)
            : m_pAllocator(pAllocator)
            , m_pGrabResult(new GrabResult())
        {
            *m_pGrabResult = grabResult;
            if (clone)
            {
                m_pGrabResult->pImageData = pAllocator->AllocateBuffer(grabResult.payloadSize);
                memcpy(m_pGrabResult->pImageData, grabResult.pImageData, grabResult.payloadSize);
            }
        }

        GrabResultPtr(GrabResultPtr& other)
            : m_pAllocator(nullptr)
            , m_pGrabResult(nullptr)
        {

            *this = other;
        }

        GrabResultPtr(Internal::GrabResultPtrRef ref)
            : m_pAllocator(ref.m_pAllocator)
            , m_pGrabResult(ref.m_pGrabResult)
        {
        }

        ~GrabResultPtr()
        {
            Reset();
        }

        GrabResultPtr& operator=(GrabResultPtr& rhs)
        {
            if (&rhs != this)
            {
                Reset(rhs.m_pGrabResult, rhs.m_pAllocator);
                rhs.Release();
            }
            return *this;
        }

        GrabResultPtr& operator=(Internal::GrabResultPtrRef ref)
        {
            if (ref.m_pGrabResult != this->Get())
            {
                Reset(ref.m_pGrabResult, ref.m_pAllocator);
            }
            return *this;
        }

        // Sets a new grabResult and allocator. The previous grab result is deleted,
        // the memory buffer of the grab result is freed, if m_pAllocator is not
        // nullptr
        void Reset(GrabResult* pGrabResult = nullptr, BufferAllocator* pAllocatpr = nullptr)
        {
            if (m_pGrabResult != nullptr && m_pGrabResult->pImageData != nullptr && m_pAllocator != nullptr)
                m_pAllocator->FreeBuffer(m_pGrabResult->pImageData);
            if (m_pGrabResult != nullptr)
                delete m_pGrabResult;
            m_pGrabResult = pGrabResult;
            m_pAllocator = pAllocatpr;
        }

        // Detaches from the GrabResult without neither deleting the GrabResult nor
        // deleting the memory the GrabResult is pointing to.
        GrabResult* Release()
        {
            GrabResult* pReturn = Get();
            m_pGrabResult = nullptr;
            m_pAllocator = nullptr;
            return pReturn;
        }

        GrabResult* Get() const { return m_pGrabResult; }

        GrabResult& operator*() const { return *Get(); }

        GrabResult* operator->() const { return Get(); }

        operator Internal::GrabResultPtrRef()
        {
            BufferAllocator *pAllocator = this->m_pAllocator;
            return Internal::GrabResultPtrRef(this->Release(), pAllocator);
        }

    private:
        BufferAllocator*        m_pAllocator;
        GrabResult*             m_pGrabResult;
    };

    enum BufferPartType
    {
        Range,
        Intensity,
        Confidence,
        Undefined
    };

    struct PartInfo
    {
        void*          pData;
        size_t         size;
        size_t         width;
        size_t         height;
        BufferPartType partType;
        uint64_t       dataFormat;
    };

    typedef std::vector<PartInfo>     BufferParts;

    class CameraList : public std::list<CameraInfo>
    {
    public:
        CameraInfo find(CameraInfoKey key, std::string value)
        {
            struct Match
            {
                Match(CameraInfoKey key, const std::string& value) : m_key(key), m_value(value) {}
                bool operator()(const CameraInfo& info) { return info.GetProperty(m_key).compare(m_value) == 0; }
            private:
                CameraInfoKey m_key;
                std::string m_value;
            };

            CameraList::const_iterator it = std::find_if(this->begin(), this->end(), Match(key, value));
            return it == cend() ? CameraInfo() : *it;
        }
    };

    class Callback;
    template<class Callee, class Camera> class RemovalCallbackT;

    template< const char ProducerName[] >
    class CGenTLCamera
    {
    public:
        typedef CProducerProxy<ProducerName> Producer_t;

        static void                  InitProducer();
        static void                  TerminateProducer();
        static bool                  IsProducerInitialized() { return s_Producer.IsIntialized(); }

        static CameraList            EnumerateCameras();

        template <class Callee, class Camera>
        void                         RegisterRemovalCallback(Callee* pClass, void (Callee::*onDeviceRemoval)(Camera* device));
        void                         UnregisterRemovalCallback();

        void                         Open(CameraInfo cameraInfo);
        void                         Open(CameraInfoKey key, const std::string& value);
        void                         OpenFirstCamera();

        bool                         IsOpen() const;
        bool                         IsConnected() const;
        void                         Close();

        CameraInfo                   GetCameraInfo() const;

        // Accessing parameters
        GenApi::INode*               GetParameter(const char* name) { return GetDeviceNodeMap()->GetNode(name); }
        const GenApi::INode*         GetParameter(const char* name) const {  return GetDeviceNodeMap()->GetNode(name); }
        GenApi::INode*               GetDeviceModuleParameter(const char* name) { return GetDeviceModuleNodeMap()->GetNode(name); }
        const GenApi::INode*         GetDeviceModuleParameter(const char* name) const { return GetDeviceModuleNodeMap()->GetNode(name); }
        GenApi::INode*               GetDataStreamParameter(const char* name) { return GetDataStreamNodeMap()->GetNode(name); }
        const GenApi::INode*         GetDataStreamParameter(const char* name) const { return GetDataStreamNodeMap()->GetNode(name); }

        void                         ParametrizeFromFile(const char* parameterFileName);

    public:

        template <typename C>
        void                         GrabContinuous(size_t nBuffers, uint32_t timeout_ms, C* pClass, bool (C::*onImageAvailable)(GrabResult, BufferParts));
        void                         GrabContinuous(size_t nBuffers, uint32_t timeout_ms, std::function< bool(GrabResult, BufferParts)> const& lambda);
        GrabResultPtr                GrabSingleImage(uint32_t timeout_ms, BufferParts* parts);
        void                         StopGrab();
        bool                         IsGrabbing() const { return 0 != (m_State & stGrabbing); }
        bool                         IsInSingleGrabMode() const { return 0 != (m_State & stSingleGrab); }
        bool                         IsAcquisitionPrepared() const { return 0 != (m_State & stAcquisitionPrepared); }

    public:

        void                         PrepareAcquisition(size_t nBuffers);
        void                         QueueBuffer(size_t idx);
        void                         QueueBuffer(GenTL::BUFFER_HANDLE hBuffer);
        void                         StartAcquisition();
        void                         StopAcquisition();
        void                         FlushQueue(GenTL::ACQ_QUEUE_TYPE operation = GenTL::ACQ_QUEUE_INPUT_TO_OUTPUT);
        void                         IssueAcquisitionStartCommand();
        void                         IssueAcquisitionStopCommand();
        void                         GetGrabResult(GenTLConsumerImplHelper::GrabResult& result, uint32_t timeout_ms);
        void                         FinishAcquisition();

        void                         SetBufferAllocator(BufferAllocator* pAllocator, bool takeOwnership);


    public:
        // Low-level functions to directly access the GenTL producer
        // Use with caution. The CGenTLCamera class doesn't notice if functions are called
        // influencing the internal state of a CGenTLCamera instance.

        // Returns an object used for directly calling the functions implemented by the producer
        static Producer_t& GetProducer() { return s_Producer; }

        // The handles for the producer's modules
        static GenTL::TL_HANDLE      GetTLHandle() { CIH_CHECK_OPEN();  return s_hTL; } 
        static GenTL::IF_HANDLE      GetInterfaceHandle() { CIH_CHECK_OPEN();  return s_hItf; } 
        GenTL::DEV_HANDLE            GetDeviceHandle() const { CIH_CHECK_OPEN();  return m_hDevice; } 
        GenTL::DS_HANDLE             GetDataStreamHandle() const { CIH_CHECK_OPEN();  return m_hDataStream; } 
        GenTL::PORT_HANDLE           GetDevicePortHandle() const { CIH_CHECK_OPEN();  return m_hDevicePort; } 

        // Get access to the node map of the device
        GenApi::INodeMap*            GetDeviceNodeMap() { CIH_CHECK_OPEN();  return m_NodeMapDevice._Ptr; }
        const GenApi::INodeMap*      GetDeviceNodeMap() const { CIH_CHECK_OPEN(); return m_NodeMapDevice._Ptr; }

        // Get access to the node map of the device module
        GenApi::INodeMap*            GetDeviceModuleNodeMap() { CIH_CHECK_OPEN(); return m_NodeMapDevModule._Ptr; }
        const GenApi::INodeMap*      GetDeviceModuleNodeMap() const { CIH_CHECK_OPEN(); return m_NodeMapDevModule._Ptr; }

        // Get access to the node map of the stream module
        GenApi::INodeMap*            GetDataStreamNodeMap() { CIH_CHECK_OPEN(); return m_NodeMapStream._Ptr; }
        const GenApi::INodeMap*      GetDataStreamNodeMap() const { CIH_CHECK_OPEN(); return m_NodeMapStream._Ptr; }

        // T is one of the described types from INFO_DATATYPE_LIST with fixed size. String types are not supported.
        template <typename T>
        void                         GetBufferInfo(GenTL::BUFFER_HANDLE hBuffer, GenTL::BUFFER_INFO_CMD infoCmd, T* value);

        // T is one of the described types from INFO_DATATYPE_LIST with fixed size, so except string types.
        template <typename T>
        void                         GetBufferPartInfo(GenTL::BUFFER_HANDLE hBuffer, uint32_t partIndex, GenTL::BUFFER_PART_INFO_CMD infoCmd, T* value);

        void                         FillGrabResult(GenTLConsumerImplHelper::GrabResult &result, const GenTL::EVENT_NEW_BUFFER_DATA &bufferData);

        void                         GetBufferParts(const GrabResult& grabResult, BufferParts& parts);

    protected:
        typedef Internal::CGrabResources<ProducerName> GrabResources;
        typedef std::unique_ptr<GrabResources>       GrabResourcesPtr;
        typedef std::unique_ptr <Internal::GrabFinalizer<ProducerName> > GrabFinalizerPtr;

    protected:
        CGenTLCamera();
        virtual ~CGenTLCamera();

        static void                 OpenTL();
        static bool                 IsTLOpen() { return IsProducerInitialized() && s_hTL != GENTL_INVALID_HANDLE; }
        static void                 CloseTL();

        static void                 OpenInterface();
        static bool                 IsInterfaceOpen() { return IsTLOpen() && s_hItf != GENTL_INVALID_HANDLE; }
        static void                 CloseInterface();

        static std::string          GetDeviceInfoString(const char* deviceId, GenTL::DEVICE_INFO_CMD cmd);


    private:
        CGenTLCamera(const CGenTLCamera&);
        CGenTLCamera& operator==(const CGenTLCamera&);

    protected:
        // Name of a parameter used for checking if a camera is still connected
        static std::string                    s_NonCachedParameter;

    private:
        static CProducerProxy<ProducerName>   s_Producer;
        static GenTL::TL_HANDLE               s_hTL;
        static GenTL::TL_HANDLE               s_hItf;
        typedef std::set<CGenTLCamera* > Cameras;
        static Cameras                        s_CamerasToClose;

    private:
        CameraInfo                            m_CameraInfo;
        GenTL::DEV_HANDLE                     m_hDevice;
        GenTL::PORT_HANDLE                    m_hDevicePort;
        GenTL::DS_HANDLE                      m_hDataStream;
        GenApi::CNodeMapRef                   m_NodeMapDevModule;
        GenApi::CNodeMapRef                   m_NodeMapDevice;
        GenApi::CNodeMapRef                   m_NodeMapStream;


    private:
        enum State
        {
            stClosed = 0,
            stOpen = 1,
            stAcquisitionPrepared = 2,
            stGrabbing = 4,
            stSingleGrab = 8
        };
        int                 m_State;

        Internal::CPort<ProducerName>                   m_DevicePort;
        Internal::CPort<ProducerName>                   m_DevModulePort;
        Internal::CPort<ProducerName>                   m_DataStreamPort;

        BufferAllocator        *m_pBufferAllocator;
        GrabResourcesPtr        m_ptrGrabResources;
        GrabFinalizerPtr        m_ptrSingleGrabFinalizer;
        bool                    m_haveAllocatorOwnership;
        Callback               *m_pRemovalCallback;
        mutable Internal::Lock  m_Lock;  // For thread safety
        WaitObject_t            m_GrabStopped;

        template<const char[]> friend struct GenTLConsumerImplHelper::Internal::GrabFinalizer;

    };

    class Callback
    {
    public:
        virtual void Fire() = 0;
        virtual ~Callback() {}
    };

    template<class Callee, class Camera>
    class RemovalCallbackT : public Callback
    {
    public:
        typedef void(Callee::*onDeviceRemoval)(Camera*);

        RemovalCallbackT(Callee* pCallee, onDeviceRemoval pMemberFunc, Camera *pCamera) : m_pCallee(pCallee), m_pMemberFunc(pMemberFunc), m_pCamera(pCamera) {}

        virtual void Fire() override
        {
            if (m_pCallee != nullptr && m_pMemberFunc != nullptr)
            {
                (m_pCallee->*m_pMemberFunc)(m_pCamera);
            }
        }
    private:
        Callee* m_pCallee;
        onDeviceRemoval m_pMemberFunc;
        Camera* m_pCamera;
    };

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Implementation of the CGenTLCamera  class
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template< const char ProducerName[] >
GenTLConsumerImplHelper::CProducerProxy<ProducerName> GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::s_Producer;

template< const char ProducerName[] >
GenTL::TL_HANDLE GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::s_hTL = GENTL_INVALID_HANDLE;

template< const char ProducerName[] >
GenTL::TL_HANDLE GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::s_hItf = GENTL_INVALID_HANDLE;

template<const char ProducerName[] >
typename GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::Cameras GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::s_CamerasToClose;

template< const char ProducerName[] >
CIH_WEAK_SYMBOL(std::string GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::s_NonCachedParameter)("DeviceTemperature");



#if defined (_MSC_VER) 
#pragma warning( push )
#   pragma warning( disable:4355 )  // 'this': used in base member initializer list
#endif
template< const char ProducerName[] >
GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::CGenTLCamera()
    : m_hDevice(GENTL_INVALID_HANDLE)
    , m_hDevicePort(GENTL_INVALID_HANDLE)
    , m_hDataStream(GENTL_INVALID_HANDLE)
    , m_State(stClosed)
    , m_DevicePort(*this)
    , m_DevModulePort(*this)
    , m_DataStreamPort(*this)
    , m_pBufferAllocator(nullptr)
    , m_haveAllocatorOwnership(false)
    , m_pRemovalCallback(nullptr)
    , m_GrabStopped(INVALID_WAIT_OBJECT)

{
    m_GrabStopped = CreateWaitObject(0);
}
#if defined (_MSC_VER) 
#   pragma warning( pop )
#endif

template< const char ProducerName[] >
GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::~CGenTLCamera()
{
    try
    {
        if (m_pRemovalCallback != nullptr)
        {
            UnregisterRemovalCallback();
        }

        if (IsOpen())
        {
            Close();  // Close will also release all dynamically allocated members
        }
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Caught exception in ~CGenTLCamera(): " << e.GetDescription() << std::endl;
    }
    CloseWaitObject(m_GrabStopped);
    SetBufferAllocator(nullptr, false); // frees the allocator if we have the ownership
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::InitProducer()
{
    if (s_Producer.IsIntialized())
        throw LOGICAL_ERROR_EXCEPTION("GenTL producer is already initialized.");

    s_Producer.Init(ProducerName);
    if (!IsTLOpen())
        OpenTL();
    OpenInterface();
}


// Doesn't throw exceptions.
template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::TerminateProducer()
{
    if (s_Producer.IsIntialized())
    {
        if (!s_CamerasToClose.empty())
        {
            Cameras camerasToClose(s_CamerasToClose);
            for (typename Cameras::iterator it = camerasToClose.begin(); it != camerasToClose.end(); ++it)
            {
                (*it)->Close();
            }
            s_CamerasToClose.clear();
        }

        if (IsInterfaceOpen())
        {
            try
            {
                CloseInterface();
            }
            catch (const GenICam::GenericException& e)
            {
                std::cerr << "Exception in TerminateProducer(): " << e.GetDescription() << std::endl;
            }
        }
        if (IsTLOpen())
        {
            try
            {
                CloseTL();
            }
            catch (const GenICam::GenericException& e)
            {
                std::cerr << "Exception in TerminateProducer():" << e.GetDescription() << std::endl;
            }
        }
    }
    s_Producer.Terminate();
}

template< const char ProducerName[] >
std::string GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetDeviceInfoString(const char* deviceId, GenTL::DEVICE_INFO_CMD cmd)
{
    CIH_CHECK_INITIALIZED();

    using namespace GenTL;

    std::vector<char> buf(1024);
    size_t bufSize = buf.size();

    INFO_DATATYPE dataType = INFO_DATATYPE_UNKNOWN;
    GC_ERROR status = s_Producer.IFGetDeviceInfo(s_hItf, deviceId, cmd, &dataType, &buf[0], &bufSize);
    if (status == GC_ERR_BUFFER_TOO_SMALL || bufSize > buf.size())
    {
        buf.resize(bufSize);
        status = s_Producer.IFGetDeviceInfo(s_hItf, deviceId, cmd, &dataType, &buf[0], &bufSize);
    }
    if (status != GC_ERR_SUCCESS)
    {
        return "<not available>";
    }
    return &buf[0];
}

template< const char ProducerName[] >
GenTLConsumerImplHelper::CameraList GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::EnumerateCameras()
{
    using namespace GenTL;

    const int DEVICE_INFO_IPADDR = DEVICE_INFO_CUSTOM_ID + 37645;  /* STRING  Query the IP address, if applicable */
    CIH_CHECK_INITIALIZED();

    GC_ERROR status;

    bool hasChanged = false;
    status = s_Producer.IFUpdateDeviceList(s_hItf, &hasChanged, 500);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to update device list"));
    }

    uint32_t numDevs = 0;
    status = s_Producer.IFGetNumDevices(s_hItf, &numDevs);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to retrieve the number of devices"));
    }

    CameraList lstCameras;

    for (uint32_t idx = 0; idx < numDevs; ++idx)
    {
        std::vector<char> buf(1024);
        size_t bufSize = buf.size();
        status = s_Producer.IFGetDeviceID(s_hItf, idx, &buf[0], &bufSize);
        if (status == GC_ERR_BUFFER_TOO_SMALL || bufSize > buf.size())
        {
            buf.resize(bufSize);
            status = s_Producer.IFGetDeviceID(s_hItf, idx, &buf[0], &bufSize);
        }
        if (status != GC_ERR_SUCCESS)
        {
            // ignore device
            std::cerr << "Failed to retrieve device ID for device " << idx << s_Producer.CreateErrorMessage("") << std::endl;
            continue;
        }
        CameraInfo info;
        info.strDeviceID = &buf[0];
        info.strDisplayName = GetDeviceInfoString(&buf[0], DEVICE_INFO_DISPLAYNAME);
        info.strModelName = GetDeviceInfoString(&buf[0], DEVICE_INFO_MODEL);
        info.strSerialNumber = GetDeviceInfoString(&buf[0], DEVICE_INFO_SERIAL_NUMBER);
        info.strUserDefinedName = GetDeviceInfoString(&buf[0], DEVICE_INFO_USER_DEFINED_NAME);
        info.strIpAddress = GetDeviceInfoString(&buf[0], DEVICE_INFO_IPADDR);
        lstCameras.push_back(info);
    }
    // Sort list by Serial number to ensure a defined order
    struct Comparer
    {
        bool operator()(const CameraInfo& left, const CameraInfo& right)
        {
            return left.strSerialNumber.compare(right.strSerialNumber) < 0;
        }
    };
    lstCameras.sort(Comparer());
    return lstCameras;
}

template< const char ProducerName[] >
template<class Callee, class Camera>
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::RegisterRemovalCallback(Callee *pCallee, void (Callee::*onDeviceRemoval)(Camera*))
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_INITIALIZED();

    if (m_pRemovalCallback != nullptr)
    {
        UnregisterRemovalCallback();
    }
    if (pCallee != nullptr)
    {
        m_pRemovalCallback = new GenTLConsumerImplHelper::RemovalCallbackT<Callee, Camera>(
            pCallee,
            onDeviceRemoval,
            static_cast<Camera*>(this));
        m_DevicePort.RegisterRemovalCallback(m_pRemovalCallback);
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::UnregisterRemovalCallback()
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_INITIALIZED();

    if (m_pRemovalCallback != nullptr)
    {
        Callback* pCallback = m_pRemovalCallback;
        m_pRemovalCallback = nullptr;
        delete pCallback;
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::Open(CameraInfo cameraInfo)
{
    Internal::LockGuard lock(m_Lock);

    using namespace GenTL;

    CIH_CHECK_INITIALIZED();

    if (IsOpen())
    {
        throw LOGICAL_ERROR_EXCEPTION("Device is already open.");
    }

    if (!cameraInfo.IsValid())
    {
        throw RUNTIME_EXCEPTION("Specified device not found or invalid camera info.");
    }

    m_CameraInfo = cameraInfo;

    GC_ERROR status;

    // open device
    {
        DEV_HANDLE hDev = GENTL_INVALID_HANDLE;
        status = s_Producer.IFOpenDevice(s_hItf, m_CameraInfo.strDeviceID.c_str(), DEVICE_ACCESS_EXCLUSIVE, &hDev);
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to open device"));
        }
        m_hDevice = hDev;
        s_CamerasToClose.insert(this);
        m_State = stOpen;
    }

    // Perform all subsequent operations in a try block to be able to clean-up properly if
    // any of the operations failed
    try
    {
        // Retrieve handle to remote device port
        {
            PORT_HANDLE hPort = GENTL_INVALID_HANDLE;
            status = s_Producer.DevGetPort(m_hDevice, &hPort);
            if (status != GC_ERR_SUCCESS)
            {
                throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to retrieve port from producer"));
            }
            m_hDevicePort = hPort;
        }

        // Create a node-map for accessing camera parameters
        bool isZipped = false;
        std::vector<char> XmlData = s_Producer.RetrieveXmlFile(GetDevicePortHandle(), &isZipped);
        try
        {
            if (isZipped)
            {
                m_NodeMapDevice._LoadXMLFromZIPData(&XmlData[0], XmlData.size());
            }
            else
            {
                m_NodeMapDevice._LoadXMLFromString(GenICam::gcstring((const char*)&XmlData[0], strlen(&XmlData[0])));  // zero termination is ensured
            }
            m_DevicePort.Attach(GetDevicePortHandle());
            m_NodeMapDevice._Connect(&m_DevicePort);
        }
        catch (GenICam::GenericException& e)
        {
            throw RUNTIME_EXCEPTION("Failed to create node map from device XML: %s", e.GetDescription());
        }

        // Create a node-map for accessing the device module's parameters
        XmlData = s_Producer.RetrieveXmlFile(GetDeviceHandle(), &isZipped);
        try
        {
            if (isZipped)
            {
                m_NodeMapDevModule._LoadXMLFromZIPData(&XmlData[0], XmlData.size());
            }
            else
            {
                m_NodeMapDevModule._LoadXMLFromString(GenICam::gcstring((const char*)&XmlData[0], strlen(&XmlData[0])));  // zero termination is ensured
            }
            m_DevModulePort.Attach(static_cast<PORT_HANDLE>(GetDeviceHandle()));
            m_NodeMapDevModule._Connect(&m_DevModulePort, "DevicePort");
        }
        catch (GenICam::GenericException& e)
        {
            throw RUNTIME_EXCEPTION("Failed to create node map from device module XML: %s", e.GetDescription());
        }

        // Open the first data stream
        uint32_t numStreams = 0;
        status = s_Producer.DevGetNumDataStreams(m_hDevice, &numStreams);
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to retrieve the number of data streams"));
        }
        if (numStreams == 0)
        {
            throw RUNTIME_EXCEPTION("Device doesn't provide a data stream.");
        }
        const uint32_t idxDataStream = 0;
        std::vector<char> bufID(512);
        size_t bufIDSize = bufID.size();
        status = s_Producer.DevGetDataStreamID(m_hDevice, idxDataStream, &bufID[0], &bufIDSize);
        if (status == GC_ERR_BUFFER_TOO_SMALL || bufIDSize > bufID.size())
        {
            bufID.resize(bufIDSize);
            status = s_Producer.DevGetDataStreamID(m_hDevice, idxDataStream, &bufID[0], &bufIDSize);
        }
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to get the ID of the first data stream"));
        }

        DS_HANDLE hDataStream = nullptr;
        status = s_Producer.DevOpenDataStream(m_hDevice, &bufID[0], &hDataStream);
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to open the first data stream"));
        }
        m_hDataStream = hDataStream;

        // Retrieve the node map for the data stream
        XmlData = s_Producer.RetrieveXmlFile(GetDataStreamHandle(), &isZipped);
        try
        {
            if (isZipped)
            {
                m_NodeMapStream._LoadXMLFromZIPData(&XmlData[0], XmlData.size());
            }
            else
            {
                m_NodeMapStream._LoadXMLFromString(GenICam::gcstring((const char*)&XmlData[0], strlen(&XmlData[0])));  // zero termination is ensured
            }
            m_DataStreamPort.Attach(static_cast<PORT_HANDLE>(GetDataStreamHandle()));
            m_NodeMapStream._Connect(&m_DataStreamPort, "StreamPort");
        }
        catch (GenICam::GenericException& e)
        {
            throw RUNTIME_EXCEPTION("Failed to create node map from data stream XML: %s", e.GetDescription());
        }
    }
    catch (...)
    {
        //clean-up state in case one of the operations failed
        try  // for safety catch exceptions, although Close() shouldn't thorw any exceptions,
        {

            Close();
        }
        catch (const GenICam::GenericException&)
        {
            // ignore exceptions that might occur
        }
        throw;
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::Open(CameraInfoKey key, const std::string& value)
{
    Internal::LockGuard lock(m_Lock);

    using namespace GenTL;

    CIH_CHECK_INITIALIZED();

    if (IsOpen())
    {
        throw LOGICAL_ERROR_EXCEPTION("Device is already open.");
    }

    Open(EnumerateCameras().find(key, value));
}

template< const char* ProducerName>
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::OpenFirstCamera()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_INITIALIZED();

    if (IsOpen())
    {
        throw LOGICAL_ERROR_EXCEPTION("Device is already open.");
    }

    CameraList lstCameras = EnumerateCameras();
    if (lstCameras.empty())
    {
        throw RUNTIME_EXCEPTION("No cameras found.");
    }
    Open(*lstCameras.begin());
}

template< const char ProducerName[] >
bool GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::IsOpen() const
{
    Internal::LockGuard lock(m_Lock);
    return s_Producer.IsIntialized() && m_State != stClosed;
}

template< const char ProducerName[] >
bool GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::IsConnected() const
{
    Internal::LockGuard lock(m_Lock);

    if (!IsOpen() || m_DevicePort.IsDisconnected())
    {
        return false;
    }

    bool retval = true;

    // The port doesn't notice a device removal until a read or write access actually have been performed.
    // Hence we perform a read access and ask the port if it has noticed a device removal. 
    if (s_NonCachedParameter.size() > 0)
    {
        try
        {
            GenApi::CValuePtr ptrValue(const_cast<GenApi::INode*>(this->GetParameter(s_NonCachedParameter.c_str())));
            assert(ptrValue.IsValid());   // parameter should be existing...
            assert(GenApi::IsReadable(ptrValue)); // ... and configured as 'readable'.
            if (GenApi::NoCache != ptrValue->GetNode()->GetCachingMode())
            {
                // The value of the node might be cached, i.e., accessing it doesn't
                // access neccessarily the hardware. Invalidate the cache in order to enforce an
                // hardware access.
                ptrValue->GetNode()->InvalidateNode();
            }
            // Read out the the value
            ptrValue->ToString();
        }
        catch (const GenICam::GenericException&)
        {
            // The register access failed. Was it due to a disconnected device?
            retval = !m_DevicePort.IsDisconnected();
        }
    }
    return retval;
}

template< const char ProducerName[] >
GenTLConsumerImplHelper::CameraInfo GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetCameraInfo() const
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();

    return m_CameraInfo;
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::Close()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_INITIALIZED();
    using namespace GenTL;

    if (!IsOpen())
    {
        std::cerr << "Close(): Device is not open." << std::endl;
    }

    if (IsGrabbing())
    {
        lock.release();
        StopGrab();
        lock.acquire(m_Lock);
    }

    if (IsAcquisitionPrepared())
    {
        FinishAcquisition(); // nothrow
    }

    GC_ERROR status;

    // device port cannot be closed. Just reset the handle.
    m_hDevicePort = GENTL_INVALID_HANDLE;

    m_NodeMapDevice._Destroy();
    m_NodeMapDevModule._Destroy();
    m_NodeMapStream._Destroy();

    // Close the data stream
    if (m_hDataStream != GENTL_INVALID_HANDLE)
    {
        status = s_Producer.DSClose(m_hDataStream);
        if (status != GC_ERR_SUCCESS)
        {
            std::cerr << s_Producer.CreateErrorMessage("Failed to close the data stream") << std::endl;
        }
        m_hDataStream = GENTL_INVALID_HANDLE;
    }

    // Close device
    assert(m_hDevice != GENTL_INVALID_HANDLE);  // Is already checked above by calling IsOpen()
    status = s_Producer.DevClose(m_hDevice);
    s_CamerasToClose.erase(this);
    m_hDevice = GENTL_INVALID_HANDLE;
    m_State = stClosed;
    if (status != GC_ERR_SUCCESS)
    {
        std::cerr << s_Producer.CreateErrorMessage("Failed to close device") << std::endl;
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::ParametrizeFromFile(const char* parameterFileName)
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();

    std::ifstream strm(parameterFileName);
    if (!strm)
    {
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(static_cast<std::stringstream&>(s << "Failed to open file " << parameterFileName).str());
    }

    GenApi::CFeatureBag FeatureBag;
    strm >> FeatureBag;
    {
        FeatureBag.LoadFromBag(GetDeviceNodeMap());

        // In some cases it is required to restore the parameters twice,
        // since on the first trial some features to restore might have been in
        // a read-only state.
        FeatureBag.LoadFromBag(GetDeviceNodeMap());
    }
}

namespace GenTLConsumerImplHelper
{
    namespace Internal {

        // Ensures proper clean-up when stopping grabbing.
        template< const char ProducerName[] >
        struct GrabFinalizer
        {
            GrabFinalizer(GenTLConsumerImplHelper::CGenTLCamera<ProducerName>& Parent)
                : acquireLock(false), m_Parent(Parent) {}

            // Destructor performs all the clean-up.
            ~GrabFinalizer()
            {
                Internal::LockGuard lock;
                if (acquireLock) {
                    lock.acquire(m_Parent.m_Lock);
                }
                m_Parent.FinishAcquisition();
                m_Parent.m_State &= ~(GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::stGrabbing | GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::stSingleGrab);
                SignalWaitObject(m_Parent.m_GrabStopped);
            }

            bool acquireLock;
        private:
            GenTLConsumerImplHelper::CGenTLCamera<ProducerName>& m_Parent;
        };
    }
}

template< const char ProducerName[] >
template< typename C >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GrabContinuous(size_t nBuffers, uint32_t timeout_ms, C* pClass, bool (C::*onImageAvailable)(GrabResult, BufferParts))
{
    using namespace GenTL;
    CIH_CHECK_OPEN();
    CIH_CHECK_NO_CONTINUOUS_GRAB();

    if (pClass == nullptr)
    {
        throw INVALID_ARGUMENT_EXCEPTION("GrabContinuous(): Pointer to class instance must not be nullptr!");
    }
    if (onImageAvailable == nullptr)
    {
        throw INVALID_ARGUMENT_EXCEPTION("GrabContinuous(): Pointer to member function must not be nullptr!");
    }

    GrabContinuous(nBuffers, timeout_ms, [&](GrabResult result, BufferParts parts) {
        return (pClass->*onImageAvailable)(result, parts);
        });
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GrabContinuous(size_t nBuffers, uint32_t timeout_ms, std::function<bool(GrabResult, BufferParts)> const& lambda)
{
    using namespace GenTL;
    CIH_CHECK_OPEN();

    if (IsInSingleGrabMode() && IsGrabbing())
    {
        StopGrab();
    }

    CIH_CHECK_NO_CONTINUOUS_GRAB();

    Internal::LockGuard lock(m_Lock);

    PrepareAcquisition(nBuffers);

    {
        Internal::GrabFinalizer<ProducerName> finalizer(*this); // Will clean-up and update status even if exceptions are thrown.

        // enqueue them to the data stream's input queue
        for (size_t i = 0; i < nBuffers; ++i)
        {
            QueueBuffer(i);
        }

        // First start the producer
        StartAcquisition();

        // ..then start the camera
        IssueAcquisitionStartCommand();

        // Acquisition loop
        bool run = true;
        BufferParts parts;
        m_State |= stGrabbing;

        finalizer.acquireLock = true;

        lock.release();

        while (run )
        {
            GrabResult result;
            GetGrabResult(result, timeout_ms);
            if (result.status == GrabResult::Ok)
            {
                GetBufferParts(result, parts);
            }
            else
            {
                parts.clear();
            }
            if (result.status != GrabResult::Stopped)
            {
                run = lambda(result, parts);
            }
            else
            {
                break;
            }

            // re-queue the buffer
            if (run && result.status != GrabResult::Timeout)
            {
                QueueBuffer(result.hBuffer);
            }
        }
    }  // The finalizer will do his work now.
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::StopGrab()
{
    using namespace GenTL;
    CIH_CHECK_OPEN();

    Internal::LockGuard lock(m_Lock);

   if (m_ptrGrabResources.get() != nullptr)
   {
       s_Producer.EventKill(m_ptrGrabResources->GetNewBufferEvent());  // wakeup anyone who might be waiting
   }
   if (IsGrabbing())
   {
       lock.release();
       WaitExWaitObject(m_GrabStopped, false);
       lock.acquire(m_Lock);
       ResetWaitObject(m_GrabStopped);
   }
   m_ptrSingleGrabFinalizer.reset();
   m_State &= ~(stGrabbing | stSingleGrab);
}


template< const char ProducerName[] >
GenTLConsumerImplHelper::GrabResultPtr GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GrabSingleImage(uint32_t timeout_ms, BufferParts* pParts)
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();

    using namespace GenTL;

    if (IsGrabbing() && !IsInSingleGrabMode())
    {
        // While a continuous grab is active, single grab is not allowed.
        throw LOGICAL_ERROR_EXCEPTION("Single grab is not allowed while continuous grab operation is in progress. Stop acquisition first.");
    }

    if (pParts != nullptr)
        pParts->resize(0);
    GrabResult result;

    assert(!IsGrabbing() || IsInSingleGrabMode());
    bool first = false;
    if (!IsGrabbing())
    {
        first = true;
        PrepareAcquisition(1);
        m_State |= (stGrabbing | stSingleGrab);
        m_ptrSingleGrabFinalizer.reset(new Internal::GrabFinalizer<ProducerName>(*this));
    }

    {
        if (!m_ptrGrabResources->GetBufferPtrs()[0].isQueued())
        {
            QueueBuffer((size_t)0);
        }

        if (first)
        {
            // First start the producer
            StartAcquisition();

            // ..then start the camera
            IssueAcquisitionStartCommand();
        }
        lock.release();
        GetGrabResult(result, timeout_ms);
        SignalWaitObject(m_GrabStopped);
        lock.acquire(m_Lock);
        if (result.status == GrabResult::Ok)
        {
            GetBufferParts(result, *pParts);
        }
    }
    return GrabResultPtr(result, m_pBufferAllocator, true);
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::FillGrabResult(GenTLConsumerImplHelper::GrabResult &result, const GenTL::EVENT_NEW_BUFFER_DATA &bufferData)
{
    using namespace GenTL;
    result = {};
    result.hBuffer = bufferData.BufferHandle;
    bool isIncomplete;
    GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_IS_INCOMPLETE, &isIncomplete);
    result.status = isIncomplete ? GrabResult::Failed : GrabResult::Ok;

    if (!isIncomplete)
    {
        GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_PAYLOADTYPE, &result.payloadType);
        GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_SIZE_FILLED, &result.payloadSize);
        if (result.payloadType != PAYLOAD_TYPE_MULTI_PART)
        {
            GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_PIXELFORMAT, &result.pixelFormat);
        }  // In case of MULTI_PART PixelFormat stays 0 to indicate that there is no pixelFormat

        GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_BASE, &result.pImageData);
        GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_FRAMEID, &result.frameID);
        GetBufferInfo(bufferData.BufferHandle, BUFFER_INFO_TIMESTAMP, &result.timeStamp);
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetBufferParts(const GrabResult& grabResult, BufferParts& parts)
{
    CIH_CHECK_INITIALIZED();
    CIH_CHECK_OPEN();

    using namespace GenTL;
    parts.resize(0);

    if (grabResult.payloadType != PAYLOAD_TYPE_MULTI_PART)
    {
        // Grab result doesn't represent a multi-part image, so return one single part
        parts.resize(1);
        parts[0].dataFormat = grabResult.pixelFormat;
        parts[0].pData = grabResult.pImageData;
        parts[0].size = grabResult.payloadSize;
        parts[0].partType = Intensity;
        GetBufferInfo(grabResult.hBuffer, BUFFER_INFO_WIDTH, &parts[0].width);
        GetBufferInfo(grabResult.hBuffer, BUFFER_INFO_HEIGHT, &parts[0].height);
        return;
    }

    uint32_t numParts;
    GC_ERROR status = s_Producer.DSGetNumBufferParts(m_hDataStream, grabResult.hBuffer, &numParts);
    if (status != GC_ERR_SUCCESS)
    {
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage(static_cast<std::ostringstream&>(s << "Failed to get number of buffer parts").str()));
    }

    parts.resize(numParts);
    for (uint32_t idx = 0; idx < numParts; ++idx)
    {
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_BASE, &parts[idx].pData);
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_DATA_SIZE, &parts[idx].size);
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_WIDTH, &parts[idx].width);
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_HEIGHT, &parts[idx].height);
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_DATA_FORMAT, &parts[idx].dataFormat);
        GenTL::PARTDATATYPE_ID partDataTypeId(PART_DATATYPE_UNKNOWN);
        GetBufferPartInfo(grabResult.hBuffer, idx, BUFFER_PART_INFO_DATA_TYPE, &partDataTypeId);

        switch (partDataTypeId)
        {
        case PART_DATATYPE_2D_IMAGE:
            // The producer is reporting the color range map as "2D image".
            if (parts[idx].dataFormat == PFNC_RGB8)
            {
                parts[idx].partType = Range;
            }
            else
            {
                parts[idx].partType = Intensity;
            }
            break;
        case PART_DATATYPE_3D_IMAGE:
            parts[idx].partType = Range;
            break;
        case PART_DATATYPE_CONFIDENCE_MAP:
            parts[idx].partType = Confidence;
            break;
        default:
            parts[idx].partType = Undefined;
        }
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::SetBufferAllocator(BufferAllocator* pAllocator, bool takeOwnership)
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_NOT_ACQ_PREPARED();  // Must not be called when acquisition is ongoing.

    if (m_pBufferAllocator && m_haveAllocatorOwnership)
    {
        m_pBufferAllocator->Delete();
    }
    m_pBufferAllocator = pAllocator;
    m_haveAllocatorOwnership = takeOwnership;
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::PrepareAcquisition(size_t nBuffers)
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();
    CIH_CHECK_NOT_ACQ_PREPARED();

    assert(m_ptrGrabResources.get() == nullptr);
    using namespace GenTL;

    if (nullptr == m_pBufferAllocator)
    {
        SetBufferAllocator(new GenTLConsumerImplHelper::HeapAllocator(), true);
    }
    GrabResourcesPtr ptrGrabResources(new GrabResources(nBuffers, *this, *m_pBufferAllocator));

    // Determine payload size
    GenApi::CIntegerPtr ptrPayloadSize(GetParameter("PayloadSize"));
    if (!ptrPayloadSize.IsValid())
    {
        throw RUNTIME_EXCEPTION("Failed to grab images: Camera doesn't provide payload size parameter.");
    }
    if (!GenApi::IsReadable(ptrPayloadSize))
    {
        throw RUNTIME_EXCEPTION("Failed to grab image. Camera's payload size is not readable.");
    }

    ptrGrabResources->SetCleanupLevel(GrabResources::FreeBuffers);
    // allocate the buffers used for grabbing
    size_t payloadSize = static_cast<size_t>(ptrPayloadSize->GetValue());
    std::vector<typename GrabResources::Buffer> &buffers = ptrGrabResources->GetBufferPtrs();
    for (size_t i = 0; i < nBuffers; ++i)
    {
        buffers[i].setBuffer(m_pBufferAllocator->AllocateBuffer(payloadSize));
    }

    ptrGrabResources->SetCleanupLevel(GrabResources::RevokeBuffers);
    // announce the buffers to the data stream
    for (size_t i = 0; i < nBuffers; ++i)
    {
        GC_ERROR status = s_Producer.DSAnnounceBuffer(GetDataStreamHandle(), buffers[i].getBuffer(), payloadSize, nullptr, &buffers[i].getHandle());
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to announce buffer"));
        }
    }

    // Register New Buffer Event. The event will be signaled if an buffer has been acquired.
    GC_ERROR status = s_Producer.GCRegisterEvent(GetDataStreamHandle(), EVENT_NEW_BUFFER, &ptrGrabResources->GetNewBufferEvent());
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to register new buffer event"));
    }
    ptrGrabResources->SetCleanupLevel(GrabResources::UnregisterGrabEvent);

    m_ptrGrabResources = std::move(ptrGrabResources);
    m_State |= stAcquisitionPrepared;

}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::QueueBuffer(GenTL::BUFFER_HANDLE hBuffer)
{
    CIH_CHECK_OPEN();

    size_t idx = m_ptrGrabResources->LookupBufferHandle(hBuffer);
    QueueBuffer(idx);
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::QueueBuffer(size_t idx)
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_OPEN();
    CIH_CHECK_ACQ_PREPARED();

    using namespace GenTL;

    typename GrabResources::Buffer& buffer = m_ptrGrabResources->GetBufferPtrs()[idx];

    if (nullptr != m_ptrGrabResources.get() && m_ptrGrabResources->GetCleanupLevel() < GrabResources::FlushQueues)
        m_ptrGrabResources->SetCleanupLevel(GrabResources::FlushQueues);

    GC_ERROR status = s_Producer.DSQueueBuffer(GetDataStreamHandle(), buffer.getHandle());
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to queue buffer"));
    }
    buffer.setQueued(true);
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::FlushQueue(GenTL::ACQ_QUEUE_TYPE operation )
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();
    CIH_CHECK_ACQ_PREPARED();

    using namespace GenTL;
    GC_ERROR status = s_Producer.DSFlushQueue(GetDataStreamHandle(), operation );
    if ( status != GC_ERR_SUCCESS )
    {
        throw RUNTIME_EXCEPTION( s_Producer.CreateErrorMessage( "Failed to flush buffers") );
    }
    
    for (auto buffer = m_ptrGrabResources->GetBufferPtrs().begin(); buffer != m_ptrGrabResources->GetBufferPtrs().end(); ++buffer)
    {
        bool isBufferQueued = false;
        GetBufferInfo( buffer->getHandle(), BUFFER_INFO_IS_ACQUIRING, &isBufferQueued );
        buffer->setQueued(isBufferQueued);
    }
}
template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::StartAcquisition()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_OPEN();
    CIH_CHECK_ACQ_PREPARED();

    using namespace GenTL;

    // Lock critical camera parameters that must not be changed while acquiring image data.
    GenApi::CIntegerPtr ptrLocked = GetParameter("TLParamsLocked");
    if (ptrLocked.IsValid() && GenApi::IsWritable(ptrLocked))
        ptrLocked->SetValue(1);

    GC_ERROR status = s_Producer.DSStartAcquisition(GetDataStreamHandle(), ACQ_START_FLAGS_DEFAULT, GENTL_INFINITE);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to start the acquisition (DSStartAcquisition failed)"));
    }
    if (nullptr != m_ptrGrabResources.get())
        m_ptrGrabResources->SetCleanupLevel(GrabResources::StopProducer);

}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::StopAcquisition()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_OPEN();
    CIH_CHECK_ACQ_PREPARED();

    using namespace GenTL;

    GC_ERROR status = s_Producer.DSStopAcquisition(GetDataStreamHandle(), ACQ_STOP_FLAGS_DEFAULT);
    if (nullptr != m_ptrGrabResources.get())
        m_ptrGrabResources->SetCleanupLevel(GrabResources::FlushQueues);

    // Lock critical camera parameters that must not be changed while acquiring image data.
    GenApi::CIntegerPtr ptrLocked = GetParameter("TLParamsLocked");
    if (ptrLocked.IsValid() && GenApi::IsWritable(ptrLocked))
        ptrLocked->SetValue(0);

    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to stop the acquisition (DSStoptAcquisition failed)"));
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::IssueAcquisitionStartCommand()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_OPEN();

    using namespace GenTL;

    GenApi::CCommandPtr ptrStartAcq = GetParameter("AcquisitionStart");
    ptrStartAcq->Execute();  // The camera now acquires images.

    if (nullptr != m_ptrGrabResources.get())
        m_ptrGrabResources->SetCleanupLevel(GrabResources::StopCamera);
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::IssueAcquisitionStopCommand()
{
    Internal::LockGuard lock(m_Lock);

    CIH_CHECK_OPEN();

    using namespace GenTL;

    GenApi::CCommandPtr ptrStartAcq = GetParameter("AcquisitionStop");
    ptrStartAcq->Execute();  // The camera now acquires images.
    if (m_ptrGrabResources.get() != nullptr && m_ptrGrabResources->GetCleanupLevel() > GrabResources::StopProducer)
        m_ptrGrabResources->SetCleanupLevel(GrabResources::StopProducer);
}


template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetGrabResult(GrabResult& result, uint32_t timeout_ms)
{
    CIH_CHECK_OPEN();

    using namespace GenTL;

    EVENT_NEW_BUFFER_DATA bufferData;

    size_t evtSize = sizeof(bufferData);
    GC_ERROR status = s_Producer.EventGetData(m_ptrGrabResources->GetNewBufferEvent(), &bufferData, &evtSize, timeout_ms);
    if (status != GC_ERR_SUCCESS)
    {
        // No data to be returned. Reset grab result.
        result.reset();
    }
    switch (status)
    {
    case GC_ERR_SUCCESS:
        m_ptrGrabResources->GetBuffer(bufferData.BufferHandle).setQueued(false);
        FillGrabResult(result, bufferData);
        break;
    case GC_ERR_TIMEOUT:
        result.status = GrabResult::Timeout;
        break;
    case GC_ERR_ABORT:
        result.status = GrabResult::Stopped;
        break;
    default:
        result.status = GrabResult::Failed;
    }
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::FinishAcquisition()
{
    Internal::LockGuard lock(m_Lock);
    CIH_CHECK_OPEN();
    CIH_CHECK_ACQ_PREPARED();
    m_ptrGrabResources.reset();
    m_State &= ~stAcquisitionPrepared;
}

// ---------------------------------------------------------------------------------------------------------------
// protected member functions
// ---------------------------------------------------------------------------------------------------------------

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::OpenTL()
{
    using namespace GenTL;

    assert(!IsTLOpen());

    if (IsTLOpen())
        throw LOGICAL_ERROR_EXCEPTION("TL cannot be opened twice.");

    // Get handle for the TL
    TL_HANDLE hTL;
    GC_ERROR status = s_Producer.TLOpen(&hTL);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to open the transport layer"));
    }
    s_hTL = hTL;
}

template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::CloseTL()
{
    using namespace GenTL;

    assert(IsTLOpen());

    if (!IsTLOpen())
        throw LOGICAL_ERROR_EXCEPTION("TL is not open.");

    GC_ERROR status = s_Producer.TLClose(s_hTL);
    s_hTL = GENTL_INVALID_HANDLE;

    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to close the transport layer"));
    }
}

// Opens the first interface provided by the producer
template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::OpenInterface()
{

    using namespace GenTL;

    assert(!IsInterfaceOpen());

    if (IsInterfaceOpen())
    {
        throw LOGICAL_ERROR_EXCEPTION("Interface cannot be opened twice.");
    }

    bool8_t bChanged = false;
    GC_ERROR status = s_Producer.TLUpdateInterfaceList(s_hTL, &bChanged, 100);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to update the interface list"));
    }

    uint32_t numInterfaces = 0;
    status = s_Producer.TLGetNumInterfaces(s_hTL, &numInterfaces);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to get number of interfaces"));
    }
    if (numInterfaces < 1)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("The producer doesn't provide any interfaces"));
    }

    const uint32_t idxInterface = 0;

    std::vector<char> bufIfaceID(1024);
    size_t iSize = bufIfaceID.size();
    status = s_Producer.TLGetInterfaceID(s_hTL, idxInterface, &bufIfaceID[0], &iSize);
    if (status == GC_ERR_BUFFER_TOO_SMALL || iSize > bufIfaceID.size())
    {
        bufIfaceID.resize(iSize);
        status = s_Producer.TLGetInterfaceID(s_hTL, idxInterface, &bufIfaceID[0], &iSize);
    }
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to retrieve the interface ID"));
    }

    IF_HANDLE hItf = GENTL_INVALID_HANDLE;
    status = s_Producer.TLOpenInterface(s_hTL, &bufIfaceID[0], &hItf);
    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to open the interface"));
    }

    s_hItf = hItf;
}


template< const char ProducerName[] >
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::CloseInterface()
{
    using namespace GenTL;

    assert(IsTLOpen() && IsInterfaceOpen());

    if (!IsTLOpen())
        throw LOGICAL_ERROR_EXCEPTION("TL is not open.");
    if (!IsInterfaceOpen())
        throw LOGICAL_ERROR_EXCEPTION("Interface is not open.");

    GC_ERROR status = s_Producer.IFClose(s_hItf);
    s_hItf = GENTL_INVALID_HANDLE;

    if (status != GC_ERR_SUCCESS)
    {
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage("Failed to close the interface"));
    }

}

template< const char ProducerName[] >
template <typename T>
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetBufferInfo(GenTL::BUFFER_HANDLE hBuffer, GenTL::BUFFER_INFO_CMD infoCmd, T* value)
{
    CIH_CHECK_INITIALIZED();
    using namespace GenTL;

    INFO_DATATYPE iType = INFO_DATATYPE_UNKNOWN;
    size_t iSize = sizeof(T);   // fixed size is used
    const GC_ERROR status = s_Producer.DSGetBufferInfo(m_hDataStream, hBuffer, infoCmd, &iType, value, &iSize);
    if (status != GC_ERR_SUCCESS)
    {
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage(static_cast<std::ostringstream&>(s << "Failed to get buffer part info (cmd=" << infoCmd << ")").str()));
    }
}

template< const char ProducerName[] >
template <typename T>
void GenTLConsumerImplHelper::CGenTLCamera<ProducerName>::GetBufferPartInfo(GenTL::BUFFER_HANDLE hBuffer, uint32_t partIndex, GenTL::BUFFER_PART_INFO_CMD infoCmd, T* pValue)
{
    CIH_CHECK_INITIALIZED();
    using namespace GenTL;

    INFO_DATATYPE dataType = INFO_DATATYPE_UNKNOWN;
    size_t size = sizeof(T);   // fixed size is used
    const GC_ERROR status = s_Producer.DSGetBufferPartInfo(m_hDataStream, hBuffer, partIndex, infoCmd, &dataType, pValue, &size);
    if (status != GC_ERR_SUCCESS)
    {
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(s_Producer.CreateErrorMessage(
            static_cast<std::ostringstream&>(s << "Failed to get buffer part info (cmd=" << infoCmd << ", partIdx=" << partIndex << ")").str()));
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Internal::CPort<CProducerName> class
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
namespace GenTLConsumerImplHelper
{
    namespace Internal
    {
        template <const char ProducerName[]>
        class CPort : public GenApi::IPort
        {
        public:
            CPort(const GenTLConsumerImplHelper::CGenTLCamera<ProducerName>& parent);
            void Attach(GenTL::PORT_HANDLE hPort);
            void Detach();
            void RegisterRemovalCallback(Callback*);
            bool IsDisconnected() const { return m_isDisconnected; }
            virtual void Read(void *pBuffer, int64_t Address, int64_t Length);
            virtual void Write(const void *pBuffer, int64_t Address, int64_t Length);
            virtual GenApi::EAccessMode GetAccessMode(void) const;
        private:
            const CGenTLCamera<ProducerName>&  m_parent;
            GenTL::PORT_HANDLE                 m_hPort;
            bool                               m_isDisconnected;
            Callback                          *m_pRemovalCallback;
        };
    }
}

template <const char ProducerName[]>
GenTLConsumerImplHelper::Internal::CPort<ProducerName>::CPort(const CGenTLCamera<ProducerName>& parent)
    : m_parent(parent)
    , m_hPort(GENTL_INVALID_HANDLE)
    , m_isDisconnected(false)
    , m_pRemovalCallback(nullptr)
{
}

template <const char ProducerName[]>
void GenTLConsumerImplHelper::Internal::CPort<ProducerName>::Attach(GenTL::PORT_HANDLE hPort)
{
    m_hPort = hPort;
    m_isDisconnected = false;
}

template <const char ProducerName[]>
void GenTLConsumerImplHelper::Internal::CPort<ProducerName>::Detach()
{
    m_hPort = GENTL_INVALID_HANDLE;
    m_isDisconnected = true;
}

template <const char ProducerName[]>
void GenTLConsumerImplHelper::Internal::CPort<ProducerName>::RegisterRemovalCallback(GenTLConsumerImplHelper::Callback* pCallback)
{
    m_pRemovalCallback = pCallback;
}

template <const char ProducerName[]>
void GenTLConsumerImplHelper::Internal::CPort<ProducerName>::Read(void *pBuffer, int64_t Address, int64_t Length)

{
    size_t size = static_cast<size_t>(Length);
    GenTL::GC_ERROR status = m_parent.GetProducer().GCReadPort(m_hPort, Address, pBuffer, &size);
    if ((status != GenTL::GC_ERR_SUCCESS) || (static_cast<int64_t>(size) != Length))
    {
        if (status == GenTL::GC_ERR_NOT_AVAILABLE)  // GC_ERR_NOT_AVAILABLE indicates device removal
        {
            if (m_pRemovalCallback != nullptr && !m_isDisconnected)
            {
                m_pRemovalCallback->Fire();
            }
            m_isDisconnected = true;
            throw RUNTIME_EXCEPTION("Device has been removed.");
        }
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(m_parent.GetProducer().CreateErrorMessage(
            static_cast<std::ostringstream&>(s << "Unable to read" << Length << "bytes from address " << std::hex << Address).str()));
    }
}

template <const char ProducerName[]>
void GenTLConsumerImplHelper::Internal::CPort<ProducerName>::Write(const void *pBuffer, int64_t Address, int64_t Length)
{
    size_t size = static_cast<size_t>(Length);
    GenTL::GC_ERROR status = m_parent.GetProducer().GCWritePort(m_hPort, Address, pBuffer, &size);
    if ((status != GenTL::GC_ERR_SUCCESS) || (static_cast<int64_t>(size) != Length))
    {
        if (status == GenTL::GC_ERR_NOT_AVAILABLE)  // GC_ERR_NOT_AVAILABLE indicates device removal
        {
            if (m_pRemovalCallback != nullptr && !m_isDisconnected)
            {
                m_pRemovalCallback->Fire();
            }
            m_isDisconnected = true;
            throw RUNTIME_EXCEPTION("Device has been removed.");
        }
        std::ostringstream s;
        throw RUNTIME_EXCEPTION(m_parent.GetProducer().CreateErrorMessage(
            static_cast<std::ostringstream&>(s << "Unable to write " << Length << "bytes to address " << std::hex << Address).str()));
    }
}

template <const char ProducerName[]>
GenApi::EAccessMode GenTLConsumerImplHelper::Internal::CPort<ProducerName>::GetAccessMode(void) const
{
    return m_parent.IsOpen() ? GenApi::RW : GenApi::NA;
}
