/*
This sample demonstrates how to calibrate a system consisting of a Basler GigE color camera
and a Basler blaze camera.

The project consists of two executables: the calibration and the fusion. In the first step,
the calibration executable must be used to determine the relative position of the cameras
to each other.

To do this, take several images of a chessboard by pressing the S key. After that, start the
calibration by pressing the C key. For the calibration 10 to 15 images of different areas
of the chessboard should be taken, covering the whole field of view. A suitable chessboard is
provided with this sample. This should be printed as DIN A3 (297 x 420 mm), so that the fields
have a side length of 40 mm each. If the side length of the fields in your printout is different,
adjust the specified size in the calibration program. Calibration only needs to be performed
once or whenever the configuration of the system is  changed. This is the case when you change,
e.g., the position of a camera.

To access the color camera, the Basler pylon Camera Software Suite (version 6 or higher) with
GigE camera support and the GigE GenTL Producer must be installed.
*/

#include <ConsumerImplHelper/BlazeCamera.h>
#include <ConsumerImplHelper/GEVCamera.h>

#include <opencv2/calib3d.hpp>
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>


using namespace GenTLConsumerImplHelper;

const std::map<std::string, cv::ColorConversionCodes> BayerFormat = {
    {"BayerGR8", cv::COLOR_BayerGR2GRAY},
    {"BayerRG8", cv::COLOR_BayerRG2GRAY},
    {"BayerBG8", cv::COLOR_BayerBG2GRAY},
    {"BayerGB8", cv::COLOR_BayerGB2GRAY}
};


class Calibration
{
public:
    int run();

private:
    void setupBlazeCamera();
    void setupColorCamera();
    cv::Mat processBlazeData(const GrabResult& result);
    cv::Mat processColorData(const GrabResult& result);
    bool locateChessboardCorners(const cv::Mat image, std::vector<cv::Point2f>& imagePoints);
    std::vector<cv::Point3f> getObjectPoints();
    void calibrateColorCamera(cv::Mat& colorCameraMatrix, cv::Mat& colorDistortion, const cv::Size imgSize);
    void getBlazeCalibration(cv::Mat& blazeCameraMatrix, cv::Mat& blazeDistortion);
    void stereoCalibration(const cv::Mat& colorCameraMatrix, const cv::Mat& colorDistortion, const cv::Mat& blazeCameraMatrix, const cv::Mat& blazeDistortion, const cv::Size imgSize);

private:
    CBlazeCamera    m_blazeCamera;
    std::string     m_bayerPattern; // The camera's color pattern.
    CGEVCamera      m_colorCamera;
    cv::Size        m_patternSize = cv::Size(8, 5); // number of corners
    float           m_fieldSize = 40.0; // mm coner size

    std::vector<std::vector<cv::Point2f>> m_blazePoints;
    std::vector<std::vector<cv::Point2f>> m_colorPoints;
    std::vector<std::vector<cv::Point3f>> m_objectPoints;

};

void Calibration::setupBlazeCamera()
{
    m_blazeCamera.Open(ModelName, "blaze-101");

    // If there are multiple cameras connected and you want to open a specific one, use
    // the CToFCamera::Open(CameraInfoKey, string) method.
    std::cout << "Connected to ToF camera " << m_blazeCamera.GetCameraInfo().strDisplayName << std::endl;

    GenApi::CEnumerationPtr ptrImageComponentSelector = m_blazeCamera.GetParameter("ComponentSelector");
    GenApi::CBooleanPtr ptrImageComponentEnable = m_blazeCamera.GetParameter("ComponentEnable");
    GenApi::CEnumerationPtr ptrPixelFormat = m_blazeCamera.GetParameter("PixelFormat");

    // Disable range data.
    ptrImageComponentSelector->FromString("Range");
    ptrImageComponentEnable->SetValue(false);

    // Enable the intensity image.
    ptrImageComponentSelector->FromString("Intensity");
    ptrImageComponentEnable->SetValue(true);
    ptrPixelFormat->FromString("Mono16");

    // Disable the confidence map.
    ptrImageComponentSelector->FromString("Confidence");
    ptrImageComponentEnable->SetValue(false);

    // Reduce exposure time to avoid overexposure at close-up range
    GenApi::CFloatPtr ptrExposureTime = m_blazeCamera.GetParameter("ExposureTime");
    ptrExposureTime->SetValue(250);

    // Configure the camera for software trigger.
    GenApi::CEnumerationPtr ptrTriggerMode = m_blazeCamera.GetParameter("TriggerMode");
    ptrTriggerMode->FromString("On");
    GenApi::CEnumerationPtr ptrTriggerSource = m_blazeCamera.GetParameter("TriggerSource");
    ptrTriggerSource->FromString("Software");
}

void Calibration::setupColorCamera()
{
    CameraList cameras = CGEVCamera::EnumerateCameras();
    std::string serialnumber = "";
    // Since the blaze camera is a GigE Vision camera, it will be enumerated too.
    // We iterate through all found cameras and open the first one which is not 
    // a blaze camera.
    for (auto cam : cameras) {
        if (cam.strModelName.find("blaze") == std::string::npos) {
            serialnumber = cam.strSerialNumber;
            break;
        }
    }
    if (serialnumber.empty())
    {
        throw RUNTIME_EXCEPTION("No 2D camera found.");
    }
    m_colorCamera.Open(SerialNumber, serialnumber);
    std::cout << "Connected to color camera " << m_colorCamera.GetCameraInfo().strDisplayName << std::endl;

    GenApi::CEnumerationPtr ptrPixelFormat = m_colorCamera.GetParameter("PixelFormat");

    // Figure out which 8-bit Bayer pixel format the camera supports.
    // If the camera supports an 8-bit Bayer format, enable the format. Otherwise, throw an
    // exception.
    m_bayerPattern = "";
    GenApi::CEnumEntryPtr ptrFormatEntry;
    for (auto const& formatName : BayerFormat) {
        ptrFormatEntry = ptrPixelFormat->GetEntryByName(formatName.first.c_str());
        if (GenApi::IsAvailable(ptrFormatEntry)) {
            m_bayerPattern = formatName.first;
            break;
        }
    }
    if (m_bayerPattern != "") {
        ptrPixelFormat->FromString(ptrFormatEntry->GetSymbolic());
    } else {
        throw RUNTIME_EXCEPTION("Camera doesn't support an 8-bit Bayer pixel format.");
    }

    // Activate auto exposure
    GenApi::CEnumerationPtr ptrExposureAuto = m_colorCamera.GetParameter("ExposureAuto");
    ptrExposureAuto->FromString("Once");

    // Configure the camera for software trigger.
    // Each software trigger will start the acquisition of one single frame.
    GenApi::CEnumerationPtr ptrAcquisitionMode = m_colorCamera.GetParameter("AcquisitionMode");
    ptrAcquisitionMode->FromString("Continuous");
    GenApi::CEnumerationPtr ptrTriggerSelector = m_colorCamera.GetParameter("TriggerSelector");
    ptrTriggerSelector->FromString("FrameStart");
    GenApi::CEnumerationPtr ptrTriggerMode = m_colorCamera.GetParameter("TriggerMode");
    ptrTriggerMode->FromString("On");
    GenApi::CEnumerationPtr ptrTriggerSource = m_colorCamera.GetParameter("TriggerSource");
    ptrTriggerSource->FromString("Software");
}

cv::Mat Calibration::processBlazeData(const GrabResult& result)
{
    BufferParts parts;
    m_blazeCamera.GetBufferParts(result, parts);

    const int width = (int)parts[0].width;
    const int height = (int)parts[0].height;
    const int count = width * height;
    cv::Mat intensity = cv::Mat(height, width, CV_16UC1, parts[0].pData);
    intensity.convertTo(intensity, CV_8UC1, 1.0 / 256.0);

    return intensity;
}

cv::Mat Calibration::processColorData(const GrabResult& result)
{
    BufferParts parts;
    m_colorCamera.GetBufferParts(result, parts);

    // The camera delivers the color information as a Bayer-coded raw image.
    // Before accessing the data, a Bayer decoding must be performed.
    cv::Mat rawImage(int(parts[0].height), int(parts[0].width), CV_8UC1, parts[0].pData);
    cv::Mat color;
    cv::cvtColor(rawImage, color, BayerFormat.at(m_bayerPattern));

    return color;
}

bool Calibration::locateChessboardCorners(const cv::Mat image, std::vector<cv::Point2f>& imagePoints)
{
    double winSize = 0;
    int flags = (cv::CALIB_CB_NORMALIZE_IMAGE		// normalize image gamma
                 | cv::CALIB_CB_FILTER_QUADS		// use additional criteria to filter out false quads at the contour retrieval stage
                 | cv::CALIB_CB_FAST_CHECK);		// run fast check on the image which looks for chessboard corners

    // Locate the chessboard corners
    bool found = cv::findChessboardCorners(image, m_patternSize, imagePoints);

    if (found) {
        // Refine the location of the corners
        cv::cornerSubPix(image, imagePoints, cv::Size(5,5), cv::Size(-1, -1), cv::TermCriteria(cv::TermCriteria::EPS + cv::TermCriteria::COUNT, 30, 0.1));
    }

    return found;
}

std::vector<cv::Point3f> Calibration::getObjectPoints()
{
    std::vector<cv::Point3f> objectPoints;
    for (int y = 0; y < m_patternSize.height; ++y) {
        for (int x = 0; x < m_patternSize.width; ++x) {
            objectPoints.push_back(cv::Point3f(x*m_fieldSize, y*m_fieldSize, 0));
        }
    }
    return objectPoints;
}

void Calibration::calibrateColorCamera(cv::Mat& colorCameraMatrix, cv::Mat& colorDistortion, const cv::Size imgSize)
{
    cv::Mat colorRotation, colorTranslation;
    const double rms = cv::calibrateCamera(
                           m_objectPoints,
                           m_colorPoints,
                           imgSize,
                           colorCameraMatrix,
                           colorDistortion,
                           colorRotation,
                           colorTranslation,
                           (cv::CALIB_FIX_ASPECT_RATIO | cv::CALIB_ZERO_TANGENT_DIST | cv::CALIB_FIX_K3));

    std::cout << "Reprojection error color camera: " << rms << std::endl;
}

void Calibration::getBlazeCalibration(cv::Mat& blazeCameraMatrix, cv::Mat& blazeDistortion)
{
    GenApi::CFloatPtr ptrScan3dFocalLength = m_blazeCamera.GetParameter("Scan3dFocalLength");
    GenApi::CFloatPtr ptrScan3dPrincipalPointU = m_blazeCamera.GetParameter("Scan3dPrincipalPointU");
    GenApi::CFloatPtr ptrScan3dPrincipalPointV = m_blazeCamera.GetParameter("Scan3dPrincipalPointV");

    blazeCameraMatrix = cv::Mat::zeros(3, 3, CV_64F);
    blazeCameraMatrix.at<double>(0, 0) = (double)ptrScan3dFocalLength->GetValue();
    blazeCameraMatrix.at<double>(0, 2) = (double)ptrScan3dPrincipalPointU->GetValue();
    blazeCameraMatrix.at<double>(1, 1) = (double)ptrScan3dFocalLength->GetValue();
    blazeCameraMatrix.at<double>(1, 2) = (double)ptrScan3dPrincipalPointV->GetValue();
    blazeCameraMatrix.at<double>(2, 2) = 1;

    blazeDistortion = cv::Mat::zeros(1, 5, CV_64F);
}

void Calibration::stereoCalibration(const cv::Mat& colorCameraMatrix, const cv::Mat& colorDistortion, const cv::Mat& blazeCameraMatrix, const cv::Mat& blazeDistortion, const cv::Size imgSize)
{
    cv::Mat rotation = cv::Mat::zeros(3, 3, CV_64F);
    cv::Mat translation = cv::Mat::zeros(1, 3, CV_64F);

    // Calibrate the Combination of RGB and 3D camera
    cv::Mat essentialMatrix, fundamentalMatrix;
    const double rms = cv::stereoCalibrate(
                           m_objectPoints,
                           m_blazePoints,
                           m_colorPoints,
                           blazeCameraMatrix,
                           blazeDistortion,
                           colorCameraMatrix,
                           colorDistortion,
                           imgSize,
                           rotation,
                           translation,
                           essentialMatrix,
                           fundamentalMatrix);

    std::cout << "Reprojection error stereo setup: " << rms << std::endl;

    // Write calibration
    cv::FileStorage fs("calibration.xml", cv::FileStorage::WRITE);
    fs << "colorCameraMatrix" << colorCameraMatrix << "colorDistortion" << colorDistortion;
    fs << "blazeCameraMatrix" << blazeCameraMatrix << "blazeDistortion" << blazeDistortion;
    fs << "rotation" << rotation << "translation" << translation;
    std::cout << "Wrote calibration to 'calibration.xml'" << std::endl;
}

int Calibration::run()
{
    int result = EXIT_SUCCESS;
    const int nBuffers = 2;   // Number of buffers used per camera
    int key = 0;

    try {
        // Open and configure the cameras.
        setupBlazeCamera();
        setupColorCamera();

        GenApi::CCommandPtr ptrColorCameraSoftwareTrigger = m_colorCamera.GetParameter("TriggerSoftware");
        GenApi::CCommandPtr ptrToFCameraSoftwareTrigger = m_blazeCamera.GetParameter("TriggerSoftware");

        // Allocate memory buffers used for grabbing.
        m_blazeCamera.PrepareAcquisition(nBuffers);
        m_colorCamera.PrepareAcquisition(nBuffers);

        unsigned int idxNextBuffer = 0;

        // For each camera, enqueue one buffer which will receive the image data of the next image
        // the corresponding camera sends.
        m_blazeCamera.QueueBuffer(idxNextBuffer);
        m_colorCamera.QueueBuffer(idxNextBuffer);

        // Start the acquisition on the host. Now, the enqueued buffers will be filled with data sent by the cameras.
        m_blazeCamera.StartAcquisition();
        m_colorCamera.StartAcquisition();

        // The camera won't send any data until they are started.
        m_blazeCamera.IssueAcquisitionStartCommand();
        m_colorCamera.IssueAcquisitionStartCommand();
        ptrColorCameraSoftwareTrigger->Execute(); // Issue software trigger for the color camera.
        ptrToFCameraSoftwareTrigger->Execute();   // Issue software trigger for the ToF camera.

        // The cameras run in software trigger mode. They will expose and transfer a new image for
        // each software trigger they receive.

        std::cout << std::endl << "Calibration of the camera system" << std::endl;
        std::cout << "  - Press 's' in one of the image windows to capture images of the chessboard for calibration" << std::endl;
        std::cout << "  - Press 'c' in one of the image windows to perform the calibration with the captured images" << std::endl;
        std::cout << "  - Press 'q' in one of the image windows to exit" << std::endl;
        std::cout << std::endl;
        std::cout << "For the calibration, about 10 to 15 images should be taken from different positions covering the entire field of view." << std::endl;

        // Grab loop. Acquire and process data from both cameras.
        // While the data is being processed for each camera, a new buffer will be filled in
        // the background.
        int cnt = 0;
        while (true) { // Process n images from each camera.
            GrabResult blazeResult;
            GrabResult colorResult;
            // Wait up to 3000 ms until the enqueued buffers have been completely filled.
            m_blazeCamera.GetGrabResult(blazeResult, 5000);
            m_colorCamera.GetGrabResult(colorResult, 5000);

            // Enqueue the next buffers.
            // The new buffers will be filled asynchronously in the background.
            // Each buffer will be filled with the next image received from the corresponding camera.
            idxNextBuffer++;
            idxNextBuffer %= nBuffers;
            m_blazeCamera.QueueBuffer(idxNextBuffer);
            m_colorCamera.QueueBuffer(idxNextBuffer);
            // Issue software triggers.
            ptrColorCameraSoftwareTrigger->Execute();
            ptrToFCameraSoftwareTrigger->Execute();

            // While the next buffers are acquired, process the current ones.
            if (blazeResult.status != GrabResult::Ok) {
                std::cerr << "Failed to grab image from ToF camera." << std::endl;
                if (blazeResult.status == GrabResult::Timeout) {
                    std::cerr << "Timeout occurred." << std::endl;
                }
                result = EXIT_FAILURE;
                break;
            }

            if (colorResult.status != GrabResult::Ok) {
                std::cerr << "Failed to grab image from color camera." << std::endl;
                if (colorResult.status == GrabResult::Timeout) {
                    std::cerr << "Timeout occurred." << std::endl;
                }
                result = EXIT_FAILURE;
                break;
            }

            // Process the data.
            cv::Mat blazeImg = processBlazeData(blazeResult);
            cv::Mat colorImg = processColorData(colorResult);

            std::vector<cv::Point2f> colorPoints;
            const bool colorFound = locateChessboardCorners(colorImg, colorPoints);
            cv::Mat colorImgResult;
            cv::cvtColor(colorImg, colorImgResult, cv::COLOR_GRAY2BGR);
            cv::drawChessboardCorners(colorImgResult, m_patternSize, colorPoints, colorFound);

            std::vector<cv::Point2f> blazePoints;
            const bool blazeFound = locateChessboardCorners(blazeImg, blazePoints);
            cv::Mat blazeImgResult;
            cv::cvtColor(blazeImg, blazeImgResult, cv::COLOR_GRAY2BGR);
            cv::drawChessboardCorners(blazeImgResult, m_patternSize, blazePoints, blazeFound);

            // Show images
            cv::imshow("blaze", blazeImgResult);
            cv::imshow("color", colorImgResult);

            char key = static_cast<char>(cv::waitKey(1));
            if ('q' == key) {
                // Cancel and close program
                break;
            } else if ('s' == key)  {
                // Saving the detected chessboard corner points
                if (blazeFound && colorFound) {
                    std::cout << "Added images from position " << cnt << std::endl;
                    m_blazePoints.push_back(blazePoints);
                    m_colorPoints.push_back(colorPoints);
                    m_objectPoints.push_back(getObjectPoints());
                    cv::imwrite("blaze_" + std::to_string(cnt) + ".png", blazeImg);
                    cv::imwrite("color_" + std::to_string(cnt) + ".png", colorImg);
                    cnt++;
                }
            } else if ('c' == key) {
                // Perform calibration using the previously captured images
                if (m_objectPoints.size() > 1) {
                    std::cout << "Calibration" << std::endl;
                    cv::Mat colorCameraMatrix, colorDistortion;
                    calibrateColorCamera(colorCameraMatrix, colorDistortion, colorImg.size());
                    cv::Mat blazeCameraMatrix, blazeDistortion;
                    getBlazeCalibration(blazeCameraMatrix, blazeDistortion);
                    stereoCalibration(colorCameraMatrix, colorDistortion, blazeCameraMatrix, blazeDistortion, colorImg.size());
                    std::cout << "Done" << std::endl;
                    break;
                } else {
                    std::cout << "Not enough images for calibration available!" << std::endl;
                }
            }
        }

        // Shut down acquisition. Stops cameras and frees all grab-related resources.
        m_blazeCamera.FinishAcquisition();
        m_colorCamera.FinishAcquisition();

        // Disable software trigger.
        GenApi::CEnumerationPtr(m_blazeCamera.GetParameter("TriggerMode"))->FromString("Off");
        GenApi::CEnumerationPtr(m_colorCamera.GetParameter("TriggerMode"))->FromString("Off");

        // Clean-up
        m_blazeCamera.Close();
        m_colorCamera.Close();
    } catch (const GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        result = EXIT_FAILURE;
    }

    return result;
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try {
        CBlazeCamera::InitProducer();
        CGEVCamera::InitProducer();

        Calibration sample;
        exitCode = sample.run();
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producers and all of their resources.
    // Note: Don't call TerminateProducer() until the destructors cameras of the camera classes
    // have been called. The destructors may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized()) {
        CBlazeCamera::TerminateProducer();
    }

    if (CGEVCamera::IsProducerInitialized()) {
        CGEVCamera::TerminateProducer();
    }

    return exitCode;
}
