# Basler blaze SDK - C++ samples

## Overview

Access to blaze cameras is provided by a GenICam-compliant GenTL producer. A
GenTL producer is a dynamic library implementing a standardized software
interface for accessing the camera.

The software interacting with the GenTL producer is called a GenTL consumer.
Using this terminology, each of the samples is a GenTL consumer, too.

The Basler blaze SDK provides the ConsumerImplHelper C++ template library that
serves as an implementation helper for GenTL consumers and thus as an API for
accessing Basler blaze cameras.

Camera configuration and access to other parameters is based on the technologies
defined by the GenICam GenApi standard. The GenApi library, the reference
implementation of the GenApi standard, is used for accessing camera parameters.

For more details how to configure a camera using the GenApi library, refer to
the 'CameraConfiguration' sample.

For more details about GenICam, GenTL, and GenApi, refer to the
<http://www.emva.org/standards-technology/genicam/> website.

## List of Samples

* FirstSample: Provides an introduction to operating a blaze camera and focuses
  on how to retreive 3D point cloud data.

* CameraConfiguration: Illustrates how to access camera parameters.

* SimpleGrabLoop: Similar to 'FirstSample'. Illustrates how to use a member
  function as grab callback.

* GrabDepthMap: Illustrates how to grab and interpret a depth map that provides
  3D information represented in a gray scale image.

* CustomGrabLoop: Illustrates how an application can provide its own grab
  loop and grab buffers.

* ConvertPointCloud2DepthMap: If you want to process 3D point clouds and depth
  maps at the same time, refer to the sample that illustrates how to extract
  depth maps from 3D point clouds. This sample also illustrates how to apply a
  rainbow color mapping to a depth map.

* ShowPointCloud: Illustrates how the Point Cloud Library (PCL) can be used to
  visualize the point clouds grabbed with a Basler blaze camera.

* ConvertAndFilter: Similar to 'ShowPointCloud'. Illustrates how to apply a
  PCL filter to 3D point clouds.

* ColorAndDepth: Illustrates how an application can trigger a camera by software
  and how to grab images from both a Basler blaze and a Basler GigE color
  camera.

* ColorAndDepthFusion: Illustrates how the data from a Basler GigE color camera
  and a Basler blaze can be fused into a colored point cloud.

* ParameterizeFromFile: Illustrates how to configure a camera by applying a
  configuration file created with the blaze Viewer.

* SavePointCloud: Illustrates the GrabSingleImage convenience method and how
  to save a point cloud to .pcd text file.

## How to Build the Samples

CMake is used as build system for building the C++ samples.

The Basler blaze SDK provides a FindBaslerBlaze.cmake module used to locate the
installation location of the Basler blaze SDK.
The find module exports the BaslerBlaze::BaslerBlazeSDK CMake target.
Pass this target as argument to target_link_libraries and CMake will take care
of setting include path, library path, etc.
Refer to the top-level CMakeLists.txt file to find out how to adjust
CMAKE_MODULE_PATH in order to ensure that CMake finds the BaslerBlaze.cmake
module.

## Windows

### Prerequisites

* CMake 3.3 or newer
* Visual Studio 2015 or newer
* Optional: OpenCV library
* Optional: Point Cloud library (PCL)
* Optional: pylon version 6 or higher with GigE camera support and the GigE
  GenTL producer

### OpenCV

Some samples depend on the OpenCV library.
To build these samples, ensure that OpenCV is installed.
You can download OpenCV here: <https://github.com/opencv/opencv/releases>

Make sure that you either download or build the OpenCV library that is
compatible with the Visual Studio version you are going to use for building
these samples.

If CMake fails to find OpenCV, set the `OpenCV_DIR` variable to the folder the
OpenCVConfig.cmake file is located in.

Example:

```cmd
   cmake -DOpenCV_DIR=C:\Program Files\opencv-3-4.1\build\x64\lib ..
```

Before you can access the `OpenCV_DIR` in the CMake-gui tool, you have to expand
the 'Ungrouped Entries' node in the CMake-gui window.

### PCL

Some samples depend on the Point Cloud Library (PCL).
When you want to build these samples, ensure that the PCL is installed.

You can download the so-called AllInOn PCL installer for Windows here:
    <https://github.com/PointCloudLibrary/pcl/releases>

If CMake fails to find the PCL, seth the `PCL_DIR` variable to the folder the
PCLConfig.cmake file is located in.

Example:
    -DPCL_DIR=C:\Program Files\PCL 1.9.1\cmake

```cmd
    cmake -DPCL_DIR=C:\Program Files\PCL 1.9.1\cmake
```

Before you can access the `PCL_DIR` in the CMake-gui tool, you have to expand
the 'Ungrouped Entries' node in the CMake-gui window.

### Building the Samples

Many C++ IDEs like Visual Studio 2017 or newer have built-in support for CMake.
Usually when using an IDE with CMake support, it is sufficient to launch the IDE
and to open the folder this document is located in and start a build.

If you want to generate Visual Studio project and solution files, you can use
CMake's CMake-gui tool.

If there is no CMake >= 3.3 installed, download a CMake Windows installer from
<https://cmake.org/download>. Launch the installer .msi and follow the
instructions.

Start CMake-gui, click the 'Browse Source...' button and navigate to the folder
this document is located in.

You have to specify a folder where the binaries will be built. Click the 'Browse
Build...' button and choose or create a folder of your choice.
It is recommended not to choose the root folder of the source tree, i.e., the
folder this document is located in. Common practice is to create a subfolder
(e.g. named `build`).

Click the 'Configure' button. A dialog will open. Choose the desired Visual
Studio version and platform. Make sure that the selected platform (x86 resp.
x64) matches the architecture the OpenCV and PCL libraries are built for.

Click the 'Finish' button to close the dialog. The configuration procedure
starts. Once the configuration procedure has succeeded, click the 'Generate'
button to create the Visual Studio project and solution files. The files will be
generated in the build folder that you specified earlier.
Navigate to that build folder and open the .sln file.

### Troubleshooting

*Symptom:*
 CMake emits lots of warnings.

*Reason:*
 Depending on the version of the PCL, the PCL's CMake configuration file causes
 many warnings.

*Solution:*
 For the purpose of building and running the Basler blaze SDK samples,
 the warnings can usually be safely ignored.

*Symptom:*
 The linker complains about a mixture of 32-bit and 64-bit libraries.

*Reason:*
 When linking an application, all binaries have to be either 32-bit or 64-bit
 binaries. Mixing 32-bit and 64-bit libraries in one executable is not
 possible.

*Solution:*
 Ensure that when generating Visual Studio solution files, you target
 the same platform (x86 vs. x64) the OpenCV und PCL libraries are built for.

## Linux

### Prerequisites

* CMake 3.3 or newer
* gcc toolchain
* Optional: development package of the OpenCV library
* Optional: development package of the Point Cloud library (PCL)
* Optional: pylon 6

On an Ubuntu system, you can install the prerequisites by issuing the following
command:

```bash
    sudo apt install -y build-essential cmake libopencv-dev libpcl-dev
```

To build the samples, open a shell, navigate to the folder this document is
located in and issue the following commands:

```bash
    mkdir build
    cd build
    cmake ..
    make -j4
```

If CMake fails to find the Basler blaze SDK, ensure that the software has been
installed correctly. If you installed the software in a location other than
/opt/Basler_blaze, use the `BaslerBlaze_ROOT_DIR` CMake variable to specify
where the Basler blaze SDK is located.

Example:

```bash
    cmake -DBaslerBlaze_ROOT_DIR=$HOME/BaslerBlaze ..
```

If CMake fails to find the OpenCV library or the PCL, make sure that the
development packages for the OpenCV libray and PCL are installed.

In case you want to use OpenCV or PCL built from sources, the `OpenCV_DIR`
and `PCL_DIR` CMake variables can be used to specify the location where the
`OpenCVConfig.cmake` resp. `PCLConfig.cmake` files are located.

Example:

```bash
    cmake -DOpenCV_DIR=$HOME/PCL/cmake -DPCL_DIR=$HOME/OpenCV/lib ..
```

If the GenTL producer is not found at program start, you need to set up your
environment as follows:

```bash
    export GENICAM_GENTL64_PATH=<BaslerBlazeRoot>/lib/gentlproducer/gtl/
```

### Troubleshooting

*Symptom:*
 On Ubuntu 16.04, the PCL samples fail to build due to a missing
 libvtkproj4.so library.

*Reason:*
 The PCL 1.7.2 from the Ubuntu 16.04 package repository has a bug. The
 CMake find-module for the PCL creates an incorrect dependency to
 libvtkproj4.so, which is not present on the system.

 *Solution:*
 Either build and install the PCL 1.8.0 or newer from sources as
 described above or provide a symbolic link to the missing library by entering
 the following command in a terminal window:
        sudo ln /usr/lib/libvtkproj4.so.5.10 /usr/lib/libvtkproj4.so

 To build the PCL from sources, download the sources from 
 https://github.com/PointCloudLibrary and build and install the PCL.
 Delete the CMakeCache.txt file in the build folder used for building
 the Basler blaze SDK samples and rerun CMake in order to use the PCL built from
 sources.

*Symptom:*
 CMake emits lots of warnings.

*Reason:*
 Depending on the version of the PCL, the PCL's CMake configuration file causes
 many warnings.

*Solution:*
 For the purpose of building and running the Basler blaze SDK samples,
 the warnings can usually be safely ignored.

## Controlling the Heartbeat Timeout

GigE Vision cameras like the blaze-101 camera require the application to
periodically access the camera by sending special network packets to the camera
in defined intervals. If the camera doesn't receive these heartbeats, it will
consider the connection as broken and won't accept any commands from the
application.

When you run your application, the GenTL producer for ToF cameras will normally
generate these heartbeats. When you set a breakpoint in your application and the
breakpoint is hit, the debugger will suspend all threads including the one
sending the heartbeats. Therefore, when you debug your application and single-
step through your code, no heartbeats are sent to the camera. As a result, the
camera notices a heartbeat timeout and closes the connection. The default value
for the heartbeat timeout is 3000 ms.

To work around this, you have to extend the heartbeat timeout during 
development. You can do this by setting an environment variable named
`GEV_HEARTBEAT_TIMEOUT` to the desired timeout in milliseconds. Using Visual
Studio, you can easily set the environment variable in the project settings for
your debugging session (in the project properties select:
Configuration Properties->Debugging->Environment).

For convenience, the GenTL producer automatically sets the heartbeat timeout to
5 minutes when running under a Visual Studio debugger. This behavior can be
overridden by setting the `GEV_HEARTBEAT_TIMEOUT` environment variable as
explained above.

When you extend the heartbeat to a high value and stop your application without
closing the device properly by calling the Close function, you won't be able to
open the camera again and will receive an error stating the device is currently
in use. This can happen if you stop your application using the debugger or if
your application terminates unexpectedly. To open the camera again you must
either wait until the timeout has elapsed or disconnect the network cable from
the camera.
